#include <iostream>
#include <glut.h>
#include <OpenGL/gl.h>
#include <stdint.h>
#include <list>
#include "glm/glm.hpp"
#include "glm/ext.hpp"

struct Vertex{
    glm::vec3 position;
    glm::vec3 normal;
    glm::vec2 texcoord;
    glm::vec4 joint;
    glm::vec4 weight;
};

struct Vbo{
    GLuint vertex;
    GLuint index;
};

struct Attribute{
    GLint handle;
    uint32_t size;
    uint64_t offset;
};

struct Shader{
    GLuint program;
    GLuint vertex;
    GLuint fragment;
    GLint mvp;
    GLint texture;
    std::list<Attribute> attributes;
};

struct Texture{
    GLuint handle;
};

struct Color{
    Color(uint8_t r = 255, uint8_t g = 255, uint8_t b = 255, uint8_t a = 255)
        : r(r), g(g), b(b), a(a) {}
    uint8_t r, g, b, a;
};

// variables
float rotation = 0;
Vbo quad;
Shader shader;
Texture texture;
const char *vertexSrc =
        "attribute vec4 position;\n"
        "attribute vec2 texcoord;\n"
        "uniform mat4 mvp;\n"
        "varying vec2 v_texcoord;\n"
        "void main(){\n"
        "   gl_Position = mvp * position;\n"
        "   v_texcoord = texcoord;\n"
        "}\n"
        "";
const char *fragmentSrc =
        "uniform sampler2D texture;\n"
        "varying vec2 v_texcoord;\n"
        "void main(){\n"
        "   gl_FragColor = texture2D(texture, v_texcoord);\n"
        "}\n"
        "";

void init(){
    glEnable(GL_DEPTH);

    // init quad buffers
    Vertex quadVertices[4];
    quadVertices[0].position = glm::vec3(-1, -1, 0); quadVertices[0].normal = glm::vec3(0, 0, -1); quadVertices[0].texcoord = glm::vec2(0, 0);
    quadVertices[1].position = glm::vec3(-1, 1, 0); quadVertices[1].normal = glm::vec3(0, 0, -1); quadVertices[1].texcoord = glm::vec2(0, 1);
    quadVertices[2].position = glm::vec3(1, 1, 0); quadVertices[2].normal = glm::vec3(0, 0, -1); quadVertices[2].texcoord = glm::vec2(1, 1);
    quadVertices[3].position = glm::vec3(1, -1, 0); quadVertices[3].normal = glm::vec3(0, 0, -1); quadVertices[3].texcoord = glm::vec2(1, 0);
    uint16_t quadIndices[6] = {0,1,2,0,2,3};

    // create quad buffers
    glGenBuffers(1, &quad.vertex);
    glBindBuffer(GL_ARRAY_BUFFER, quad.vertex);
    glBufferData(GL_ARRAY_BUFFER, 4 * sizeof(Vertex), &quadVertices[0], GL_STATIC_DRAW);
    glGenBuffers(1, &quad.index);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, quad.index);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, 6 * sizeof(uint16_t), &quadIndices[0], GL_STATIC_DRAW);

    // create shader
    shader.vertex = glCreateShader(GL_VERTEX_SHADER);
    shader.fragment = glCreateShader(GL_FRAGMENT_SHADER);
    int vertexSrcLen = strlen(vertexSrc);
    int fragmentSrcLen = strlen(fragmentSrc);
    glShaderSource(shader.vertex, 1, &vertexSrc, &vertexSrcLen);
    glShaderSource(shader.fragment, 1, &fragmentSrc, &fragmentSrcLen);
    glCompileShader(shader.vertex);
    glCompileShader(shader.fragment);
    shader.program = glCreateProgram();
    glAttachShader(shader.program, shader.vertex);
    glAttachShader(shader.program, shader.fragment);
    glLinkProgram(shader.program);

    // init shader
    Attribute positionAttrib;
    positionAttrib.handle = glGetAttribLocation(shader.program, "position");
    positionAttrib.offset = 0;
    positionAttrib.size = 3;
    shader.attributes.push_back(positionAttrib);
    Attribute texcoordAttrib;
    texcoordAttrib.handle = glGetAttribLocation(shader.program, "texcoord");
    texcoordAttrib.offset = 24;
    texcoordAttrib.size = 2;
    shader.attributes.push_back(texcoordAttrib);
    shader.mvp = glGetUniformLocation(shader.program, "mvp");
    shader.texture = glGetUniformLocation(shader.program, "texture");

    // create texture
    Color bitmap[32 * 32];
    for (uint y = 0; y < 32; ++y)
        for (uint x = 0; x < 32; ++x){
            bitmap[y * 32 + x] = Color(float(x) / 32.0f * 255, float(y) / 32.0f * 255, 1.0 * 255, 1.0 * 255);
        }
    glGenTextures(1, &texture.handle);
    glBindTexture(GL_TEXTURE_2D, texture.handle);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 32, 32, 0, GL_RGBA, GL_UNSIGNED_BYTE, bitmap);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glGenerateMipmap(GL_TEXTURE_2D);
}

void render(){
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // bind vbo
    glBindBuffer(GL_ARRAY_BUFFER, quad.vertex);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, quad.index);

    // prepare shader
    glUseProgram(shader.program);
    glm::mat4 mvp = glm::perspective(45.0f, 800.0f / 480.0f, 1.0f, 100.0f) * glm::lookAt(glm::vec3(0, 0, -5), glm::vec3(0.0), glm::vec3(0, 1, 0)) * glm::eulerAngleYXZ(rotation+=0.01, 0.0f, 0.0f);
    glUniformMatrix4fv(shader.mvp, 1, GL_FALSE, &mvp[0][0]);
    glBindTexture(GL_TEXTURE_2D, texture.handle);
    glActiveTexture(GL_TEXTURE0);
    glUniform1i(shader.texture, 0);

    // prepare attributes
    for (std::list<Attribute>::iterator it = shader.attributes.begin(), end = shader.attributes.end(); it != end; ++it){
        glEnableVertexAttribArray(it->handle);
        glVertexAttribPointer(it->handle, it->size, GL_FLOAT, GL_FALSE, sizeof(Vertex), reinterpret_cast<void*>(it->offset));
    }

    glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_SHORT, 0);

    for (std::list<Attribute>::iterator it = shader.attributes.begin(), end = shader.attributes.end(); it != end; ++it)
        glDisableVertexAttribArray(it->handle);

    glutSwapBuffers();
}

void reshape(int w, int h){
    glViewport(0, 0, w, h);
}

void cleanup(){
    glDeleteBuffers(1, &quad.vertex);
    glDeleteBuffers(1, &quad.index);
}

int main(int argc, char **argv)
{
    glutInit(&argc, argv);
    glutInitWindowSize(800, 480);
    glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
    glutCreateWindow("Test");

    init();

    std::cout << "Vertex size " << sizeof(Vertex) << " bytes" << std::endl;
    std::cout << "Quad: vertex " << quad.vertex << " index " << quad.index << std::endl;

    glutDisplayFunc(render);
    glutReshapeFunc(reshape);
    glutIdleFunc(render);

    glutMainLoop();

    cleanup();
    return 0;
}

